/* Implementation of class Game.
 *
 * Copyright (c) 2003 Ole Laursen.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

#include <config.h>
#include <algorithm>
#include <cassert>
#include <gdk/gdkkeysyms.h>

#include "game.hpp"

#include "main-window.hpp"
#include "image-cache.hpp"
#include "player.hpp"
#include "hero.hpp"
#include "monsters.hpp"
#include "obstacles.hpp"
#include "fading-decoration.hpp"
#include "persistent.hpp"

#include "i18n.hpp"
#include "ucompose.hpp"


Game *Game::current_instance = 0;

Game::Game(int start_level, Vector<int> dim)
  : arena(dim), level(start_level)
{
  current_instance = this;

  // setup size
  Vector<int> real_dim((dim.x + 1) * Arena::tile_width,
		       (dim.y + 1) * Arena::tile_height);

  // offset the region so that we get to see only half of each of the outer
  // wall boundary tiles
  Vector<int> offset(Arena::tile_width / 2, Arena::tile_height / 2);
  MainWindow::instance().set_canvas_size(real_dim, offset);

  Persistent &per = Persistent::instance();
    
  // setup players
  for (int i = 1; i <= per.no_of_players; ++i) {
    Player *p = new Player(i);
    players.push_back(p);
  }

  if (level > 1) {
    Persistent::LevelInfo &li
      = per.state[per.no_of_players-1][per.difficulty_level].levels[level - 2];
    
    Player::resurrections = li.resurrections;

    for (int i = 0; i < per.no_of_players; ++i)
      players[i]->set_score(li.scores[i]); 
  }
  else if (level == 1)
    Player::resurrections = 3 * per.no_of_players;
  
  init_level();

  // and action!
  Glib::signal_timeout()
    .connect(SigC::slot(*this, &Game::main_loop), 1000 / iterations_per_sec);
}

Game::~Game()
{
  clear_arena();
  
  for (player_sequence::iterator i = players.begin(),
	 end = players.end(); i != end; ++i)
    delete *i;
}

Game &Game::instance()
{
  assert(current_instance != 0);

  return *current_instance;
}

bool Game::main_loop()
{
  switch (state) {
  case before_level:
    if (--state_countdown < 0) {
      MainWindow::instance().remove_message();
      state = playing;
    }
    break;
    
  case playing:
    process_round();
    if (monsters.empty() && eggs.empty()) {
      MainWindow::instance()
	.display_message(String::ucompose(_("Level %1 cleared for monsters!"),
					  level));
      
      // update state
      Persistent::LevelInfo li;
      li.resurrections = Player::resurrections;
      li.scores.reserve(players.size());
      
      for (player_sequence::size_type i = 0; i < players.size(); ++i)
	li.scores.push_back(players[i]->get_score());

      Persistent &p = Persistent::instance();
      Persistent::StateInfo &pi
	= p.state[p.no_of_players-1][p.difficulty_level];
	
      if (int(pi.levels.size()) < level)
	pi.levels.push_back(li);
      else
	pi.levels[level - 1] = li;
	
      if (level + 1 <= max_levels)
	pi.last = level + 1;
      else
	pi.last = max_levels;
      
      Persistent::instance().save_state();
	
      state_countdown = iterations_per_sec;
      state = passed_level;
    }
    else if (!players_alive()) {
      MainWindow::instance()
	.display_message(_("And the last hero disappeared..."));
      state_countdown = iterations_per_sec * 2; 	
      state = lost_game;
    }
    
    break;
    
  case paused:
    break;

  case passed_level:
    if (--state_countdown < 0) {
      if (++level == max_levels + 1) {
	MainWindow::instance()
	  .display_message(_("All levels liberated!"));
	state_countdown = iterations_per_sec * 3;
	state = won_game;
      }
      else
	init_level();
    }
    else
      process_round();
    break;

  case won_game:
    if (--state_countdown < 0)
      MainWindow::instance().end_game();
    else
      process_round();
    break;

  case lost_game:
    if (--state_countdown < 0)
      MainWindow::instance().end_game();
    else
      process_round();
    break;
    
  default:
    break;
  }
  
  return true;			// always continue
}

void Game::key_pressed(unsigned int keyval)
{
#if 0 // cheating, unsafe
  if (keyval == GDK_F5) {
    ++level;
    init_level();
  }
#endif
  
  if (state == paused) {
    if (keyval == GDK_space) {
      MainWindow::instance().unpause_game();
      state = playing;
    }
  }
  else
    for (std::vector<Player *>::iterator i = players.begin(),
	   end = players.end(); i != end; ++i)
      (*i)->key_pressed(keyval);
}

void Game::key_released(unsigned int keyval)
{
  for (std::vector<Player *>::iterator i = players.begin(),
	 end = players.end(); i != end; ++i)
    (*i)->key_released(keyval);
}

void Game::add_decoration(FadingDecoration *deco, Block *block)
{
  decos.push_back(std::make_pair(deco, block));
}

void Game::add_monster(Monster *m)
{
  monsters.push_back(m);
}
  
void Game::remove_monster(Monster *monster)
{
  monster_sequence::iterator i
    = find(monsters.begin(), monsters.end(), monster);

  assert(i != monsters.end());

  monsters.erase(i);

  delete monster;
}

void Game::generate_egg(Vector<int> pos)
{
  Egg *egg;
  int sum = plain_monsters + hard_monsters
    + seeker_monsters + egg_layer_monsters;
  int no = std::rand() % sum;

  if (no < plain_monsters)
    egg = new PlainEgg(pos);
  else if (no < plain_monsters + hard_monsters)
    egg = new HardEgg(pos);
  else if (no < plain_monsters + hard_monsters + seeker_monsters)
    egg = new SeekerEgg(pos);
  else
    egg = new EggLayerEgg(pos);

  eggs.push_back(egg);
}
  
void Game::remove_egg(Egg *egg)
{
  egg_sequence::iterator i = find(eggs.begin(), eggs.end(), egg);

  assert(i != eggs.end());

  eggs.erase(i);

  delete egg;
}

bool Game::pause()
{
  if (state == playing)
    state = paused;

  return state == paused;
}

// return value of x given two points on a straight line
// inline float interp(float x, int x1, float y1, int x2, float y2)
// {
//   return (y2 - y1) / (x2 - x1) * (x - x1) + y1;
// }

void Game::init_level()
{
  clear_arena();

  // reward players
  if (level > 1)
    for (player_sequence::iterator i = players.begin(),
	   end = players.end(); i != end; ++i)
      if ((*i)->get_hero())
	(*i)->scored(10);

  Vector<int> dim = arena.get_dim();
  int tiles = dim.x * dim.y;
  
  create_outer_walls();
  create_obstacles(int(0.03 * tiles), int(0.35 * tiles));

  //                  lvl pln hrd sek egg-layer
  batch_create_monsters(1, 12, 0, 0, 0);
  batch_create_monsters(2, 16, 0, 0, 0);
  batch_create_monsters(3, 20, 0, 0, 0);
  batch_create_monsters(4, 24, 0, 0, 0);
  batch_create_monsters(5, 28, 1, 0, 0);
  batch_create_monsters(6, 15, 2, 0, 0);
  batch_create_monsters(7, 15, 4, 0, 0);
  batch_create_monsters(8, 15, 6, 0, 0);
  batch_create_monsters(9, 15, 8, 0, 0);
  batch_create_monsters(10, 20, 0, 1, 0);
  batch_create_monsters(11, 20, 0, 3, 0);
  batch_create_monsters(12, 20, 1, 5, 0);
  batch_create_monsters(13, 20, 2, 6, 0);
  batch_create_monsters(14, 20, 2, 8, 0);
  batch_create_monsters(15, 20, 0, 2, 1);
  batch_create_monsters(16, 20, 0, 4, 2);
  batch_create_monsters(17, 20, 0, 4, 3);
  batch_create_monsters(18, 20, 0, 4, 4);
  batch_create_monsters(19, 20, 0, 4, 5);
  batch_create_monsters(20, 20, 1, 4, 4);
  batch_create_monsters(21, 25, 1, 4, 4);
  batch_create_monsters(22, 30, 1, 4, 4);
  batch_create_monsters(23, 35, 1, 4, 4);
  batch_create_monsters(24, 25, 3, 4, 4);
  batch_create_monsters(25, 20, 5, 4, 4);
  batch_create_monsters(26, 20, 8, 4, 5);
  batch_create_monsters(27, 20, 8, 6, 6);
  batch_create_monsters(28, 20, 8, 8, 7);
  batch_create_monsters(29, 20, 8, 10, 8);
  batch_create_monsters(30, 20, 8, 12, 12);

  for (player_sequence::iterator i = players.begin(),
	 end = players.end(); i != end; ++i)
    if (!(*i)->is_dead())
      (*i)->spawn_hero();

  MainWindow::instance().set_level_info(level);
  MainWindow::instance()
    .display_message(String::ucompose(_("Entering level %1..."), level));

  state_countdown = iterations_per_sec;
  state = before_level;
}

void Game::create_outer_walls()
{
  Vector<int> dim = arena.get_dim();
  
  Wall *w = 0;

  // build the walls around the arena
  for (int x = 1; x < dim.x - 1; ++x) {
    w = new Wall;
    w->occupy_tile(make_vector(x, 0));
    walls.push_back(w);

    w = new Wall;
    w->occupy_tile(make_vector(x, dim.y - 1));
    walls.push_back(w);
  }

  for (int y = 0; y < dim.y; ++y) {
    w = new Wall;
    w->occupy_tile(make_vector(0, y));
    walls.push_back(w);

    w = new Wall;
    w->occupy_tile(make_vector(dim.x - 1, y));
    walls.push_back(w);
  }
}

void Game::create_obstacles(int no_walls, int no_blocks)
{
  walls.reserve(walls.size() + no_walls);
  blocks.reserve(blocks.size() + no_blocks);
  
  // place random walls
  for (int i = 0; i < no_walls; ++i) {
    Wall *w = new Wall;
    w->occupy_tile(arena.random_tile());
    walls.push_back(w);
  }

  // and random blocks
  for (int i = 0; i < no_blocks; ++i) {
    Block *b = new Block;
    b->occupy_tile(arena.random_tile());
    blocks.push_back(b);
  }
}


template <class MonsterType> void Game::create_monsters(int no)
{
  for (int i = 0; i < no; ++i) {
    Monster *m = new MonsterType;
    m->occupy_tile(arena.random_tile());
    add_monster(m);
  }
}

void Game::batch_create_monsters(int l, int plain, int hard, int seeker,
				 int eggl)
{
  if (level == l) {
    plain_monsters = plain;
    create_monsters<PlainMonster>(plain);
    hard_monsters = hard;
    create_monsters<HardMonster>(hard);
    seeker_monsters = seeker;
    create_monsters<SeekerMonster>(seeker);
    egg_layer_monsters = eggl;
    create_monsters<EggLayerMonster>(eggl);
  }
}


void Game::process_round()
{
  // service monsters
  for (monster_sequence::iterator i = monsters.begin(),
	 end = monsters.end(); i != end; ++i)
    (*i)->update();

  // service eggs - be careful, we might have to remove one
  for (egg_sequence::iterator i = eggs.begin(),
	 end = eggs.end(); i != end; ) {
    Egg *egg = *(i++);
    if (egg->hatch_ready()) {
      egg->hatch();
      remove_egg(egg);
    }
    else
      egg->update();
  }
  

  // service heros
  MainWindow::instance().set_resurrections(Player::resurrections);
  int no = 1;
  for (player_sequence::iterator i = players.begin(),
	 end = players.end(); i != end; ++i) {
    Player *p = *i;
    p->update();
    MainWindow::instance().set_player_info(no++, p->get_score());
  }

  // service decorations
  for (deco_sequence::iterator i = decos.begin(),
	 end = decos.end(); i != end;) {
    FadingDecoration *d = i->first;

    d->update();

    if (d->last_frame()) {
      if (Block *block = i->second)
	block->remove_cargo(&d->get_graphic());

      delete d;
      decos.erase(i++);
    }
    else
      ++i;
  }
}

bool Game::players_alive()
{
  for (player_sequence::iterator i = players.begin(),
	 end = players.end(); i != end; ++i)
    if (!(*i)->is_dead())
      return true;

  return false;
}

void Game::clear_arena()
{
  for (deco_sequence::iterator i = decos.begin(),
	 end = decos.end(); i != end; ++i) {
    FadingDecoration *d = i->first;
    if (Block *block = i->second)
      block->remove_cargo(&d->get_graphic());
    delete d;
  }
  decos.clear();
  
  for (player_sequence::iterator i = players.begin(),
	 end = players.end(); i != end; ++i)
    (*i)->clear_hero();
  
  for (monster_sequence::iterator i = monsters.begin(),
	 end = monsters.end(); i != end; ++i)
    delete *i;
  monsters.clear();

  for (egg_sequence::iterator i = eggs.begin(),
	 end = eggs.end(); i != end; ++i)
    delete *i;
  eggs.clear();
  
  for (wall_sequence::iterator i = walls.begin(),
	 end = walls.end(); i != end; ++i)
    delete *i;
  walls.clear();

  for (block_sequence::iterator i = blocks.begin(),
	 end = blocks.end(); i != end; ++i)
    delete *i;
  blocks.clear();

}

