/* Implementation of Hero class.
 *
 * Copyright (c) 2003 Ole Laursen.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

#include <cassert>

#include "hero.hpp"

#include "arena.hpp"
#include "main-window.hpp"
#include "game.hpp"
#include "graphic.hpp"
#include "image-cache.hpp"
#include "monsters.hpp"
#include "obstacles.hpp"
#include "player.hpp"

int const hero_move_rounds = 6;

Hero::Hero(Player &p)
  : Thing(new Graphic(MainWindow::instance().get_creature_layer())),
    movement(*this, hero_move_rounds), player(p), queued_move_dir(0, 0)
{
  get_graphic().set_pixbuf(ImageCache::instance().get("hero.png").get(0));
}

#if 0
#include <fstream>		// FIXME: delete this
std::ofstream aos("arena");

void dump_arena()
{
  Vector<int> v = Arena::instance().get_dim();
    
  for (int y = 0; y < v.y; ++y) {
    for (int x = 0; x < v.x; ++x) {
      Thing *t = Arena::instance().occupier(make_vector(x, y));
      if (!t)
	aos << ' ';
      else if (dynamic_cast<Monster *>(t))
	aos << 'm';
      else if (dynamic_cast<Wall *>(t))
	aos << 'W';
      else if (dynamic_cast<Block *>(t))
	aos << 'B';
      else if (dynamic_cast<Hero *>(t))
	aos << 'H';
    }
    aos << std::endl;
  }
}
#endif

void Hero::update()
{
  if (monster_encountered()) {
    player.kill_hero();
    // FIXME: splat 
    return;
  }
  
  if (movement.is_moving()) {
    bool halfway = movement.is_halfway();
    
    if (!movement.has_moved())
      if (!valid_dir(move_dir))
	movement.turn_around();
      else if (halfway)
	possibly_push_blocks(move_dir);
    
    movement.update();

    // first pull when we've moved
    if (halfway && pull && pullable_block(tile_pos - 2 * move_dir))
      pull_block(tile_pos - 2 * move_dir, move_dir);
  }
  else {
    if (tile_pos + queued_move_dir != tile_pos && valid_dir(queued_move_dir)) {
      move_dir = queued_move_dir;
      pull = queued_pull;
      movement.head(move_dir);
      // FIXME: perhaps show that we are pulling/pushing
    }
    
    queued_move_dir = make_vector(0, 0);
  }
}

void Hero::order_move(Vector<int> dir, bool d)
{
  queued_move_dir = dir;
  queued_pull = d;
}

bool Hero::valid_dir(Vector<int> dir)
{
  Vector<int> tile = tile_pos + dir;
  Thing *thing = Arena::instance().occupier(tile);

  if (!thing)
    return true;
  else if (dynamic_cast<Block *>(thing)) {
    // consider whether we can push this block
    for (;;) {
      tile += dir;
      thing = Arena::instance().occupier(tile);
      
      if (!thing)
	return true;
      else if (dynamic_cast<Hero *>(thing))
	return false;
      else if (dynamic_cast<Wall *>(thing))
	return false;
      else if (Monster *m = dynamic_cast<Monster *>(thing)) {
	// check whether next tile is good for splatting
	tile += dir;
	thing = Arena::instance().occupier(tile);
	
	if (dynamic_cast<Wall *>(thing) ||
	    dynamic_cast<Block *>(thing) && !m->hard)
	  return true;
	else
	  return false;
      }
      else if (dynamic_cast<Egg *>(thing))
	return true;		// we can always splat eggs directly
    }
  }
  else
    return false;
}

#include <libgnome/gnome-sound.h> // FIXME

void Hero::possibly_push_blocks(Vector<int> dir)
{
  Vector<int> tile = tile_pos + dir;
  Block *block = dynamic_cast<Block *>(Arena::instance().occupier(tile));

  // no blocks anyway
  if (!block)
    return;

  // else move each encountered block
  Thing *next;
  do {
    tile += dir;
    next = Arena::instance().occupier(tile);
    block->move_block_to(tile);
    block = dynamic_cast<Block *>(next);
  } while (block);

  if (Monster *monster = dynamic_cast<Monster *>(next)) {
    // we splatted a monster
    monster->splat(dir);
    player.scored(monster->kill_score);

    Game::instance().remove_monster(monster);
  }
  else if (Egg *egg = dynamic_cast<Egg *>(next)) {
    // we splatted an egg
    egg->splat(dir);
    player.scored(Egg::kill_score);
    
    Game::instance().remove_egg(egg);
  }

  //gnome_sound_play("../sounds/clinck.wav");
}

bool Hero::pullable_block(Vector<int> tile)
{
  Arena &arena = Arena::instance();
  
  // it needs to be a block
  if (!dynamic_cast<Block *>(arena.occupier(tile)))
    return false;

  // we don't allow monsters around the block, then it would be too easy to
  // splat them by repeatedly pulling and pushing
  if (dynamic_cast<Monster *>(arena.occupier(tile + make_vector(-1,-1))) ||
      dynamic_cast<Monster *>(arena.occupier(tile + make_vector( 0,-1))) ||
      dynamic_cast<Monster *>(arena.occupier(tile + make_vector( 1,-1))) ||
      dynamic_cast<Monster *>(arena.occupier(tile + make_vector(-1, 0))) ||
      dynamic_cast<Monster *>(arena.occupier(tile + make_vector( 1, 0))) ||
      dynamic_cast<Monster *>(arena.occupier(tile + make_vector(-1, 1))) ||
      dynamic_cast<Monster *>(arena.occupier(tile + make_vector( 0, 1))) ||
      dynamic_cast<Monster *>(arena.occupier(tile + make_vector( 1, 1))))
    return false;
  else
    return true;
}

void Hero::pull_block(Vector<int> tile, Vector<int> dir)
{
  Block *block = dynamic_cast<Block *>(Arena::instance().occupier(tile));
  assert(block != 0);

  // FIXME: this looks "jumpy"
  Arena::instance().occupy(tile, 0);
  block->move_block_to(tile + dir);
}

Monster *Hero::monster_encountered()
{
  for (Game::monster_sequence::iterator i = Game::instance().monsters.begin(),
	 end = Game::instance().monsters.end(); i != end; ++i)
    // the collision detection is rather crude, simply distance
    if (dist((*i)->get_graphic().get_pos(), get_graphic().get_pos())
	< (Arena::tile_width + Arena::tile_height) / 4)
      return *i;

  return 0;
}
