/* Implementation of MainWindow.
 *
 * Copyright (c) 2003 Ole Laursen.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

#include <config.h>

#include <cassert>

#include <gtkmm/main.h>

#include "main-window.hpp"

#include "preferences-window.hpp"
#include "new-game-window.hpp"
#include "intro.hpp"
#include "game.hpp"
#include "canvas-message.hpp"
#include "graphic.hpp"
#include "pixbuf-drawing.hpp"

#include "i18n.hpp"
#include "gui-helpers.hpp"
#include "ucompose.hpp"

MainWindow &MainWindow::instance()
{
  static MainWindow w;
  return w;
}

MainWindow::MainWindow()
  : gconf_dir("/apps/monster-masher")
{
  gconf_client = Gnome::Conf::Client::get_default_client();
  gconf_client->add_dir(gconf_dir);
  
  ui = get_glade_xml("main_window");
  ui->get_widget("main_window", window);

  ui->get_widget("resurrections_label", resurrections_label);
  ui->get_widget("player1_status_label", player1_status_label);
  ui->get_widget("player2_status_label", player2_status_label);
  ui->get_widget("level_label", level_label);

  resurrections_label->set_text("");
  player1_status_label->set_text("");
  player2_status_label->set_text("");
  level_label->set_text("");
  
  
  // setup canvas
  ui->get_widget("canvas", canvas);

  canvas->property_aa() = true;

  Gnome::Canvas::Group &root = *canvas->root();

  background_layer = manage(new Gnome::Canvas::Group(root));
  block_layer = manage(new Gnome::Canvas::Group(root));
  creature_layer = manage(new Gnome::Canvas::Group(root));
  text_layer = manage(new Gnome::Canvas::Group(root));

  set_canvas_size(make_vector(600, 400), make_vector(0, 0));
  intro.reset(new Intro);

  
  // connect menus
  Gtk::MenuItem *menu_item;

  ui->get_widget("new_game_item", new_game_item);
  new_game_item->signal_activate()
    .connect(SigC::slot(*this, &MainWindow::on_new_game_activated));
  
  ui->get_widget("end_game_item", end_game_item);
  end_game_item->signal_activate()
    .connect(SigC::slot(*this, &MainWindow::end_game));
  
  ui->get_widget("pause_game_item", pause_game_item);
  pause_game_item->signal_activate()
    .connect(SigC::slot(*this, &MainWindow::on_pause_game_activated));

  ui->get_widget("introduction_item", introduction_item);
  introduction_item->signal_activate()
    .connect(SigC::slot(*this, &MainWindow::on_introduction_activated));

  // no game started yet
  end_game_item->set_sensitive(false);
  pause_game_item->set_sensitive(false);
  introduction_item->set_sensitive(true);
  
  ui->get_widget("preferences_item", menu_item);
  menu_item->signal_activate()
    .connect(SigC::slot(*this, &MainWindow::on_preferences_activated));
  
  ui->get_widget("quit_item", menu_item);
  menu_item->signal_activate()
    .connect(SigC::slot(*this, &MainWindow::on_quit_activated));
  
  ui->get_widget("about_item", menu_item);
  menu_item->signal_activate()
    .connect(SigC::slot(*this, &MainWindow::on_about_activated));


  // connect various
  window->signal_key_press_event()
    .connect_notify(SigC::slot(*this, &MainWindow::on_key_press_event));
  window->signal_key_release_event()
    .connect_notify(SigC::slot(*this, &MainWindow::on_key_release_event));

  // connect window closement
  window->signal_delete_event()
    .connect(SigC::slot(*this, &MainWindow::on_closed));

  window->show_all();
}

MainWindow::~MainWindow()
{
  // make sure intro doesn't try to hide message when we are dead
  intro.reset();
}
  
Gtk::Window &MainWindow::get_window()
{
  return *window;
}

Gnome::Canvas::Group &MainWindow::get_background_layer()
{
  return *background_layer;
}

Gnome::Canvas::Group &MainWindow::get_block_layer()
{
  return *block_layer;
}

Gnome::Canvas::Group &MainWindow::get_creature_layer()
{
  return *creature_layer;
}

Gnome::Canvas::Group &MainWindow::get_text_layer()
{
  return *text_layer;
}


void MainWindow::set_canvas_size(Vector<int> dim, Vector<int> offset)
{
  canvas->set_scroll_region(offset.x, offset.y,
			    dim.x + offset.x, dim.y + offset.y);
  canvas->set_size_request(dim.x, dim.y);
  // the window will actually be larger than this because of menus etc.
  window->resize(dim.x, dim.y);
  
  generate_background(dim);
}

void MainWindow::set_resurrections(int resurrections)
{
  resurrections_label->set_text(String::ucompose("Resurrections: %1",
						 resurrections));
}

void MainWindow::set_player_info(int player_no, int score)
{
  Glib::ustring s
    = String::ucompose("Score: %1", score);
  
  if (player_no == 1)
    player1_status_label->set_text(s);
  else
    player2_status_label->set_text(s);
}

void MainWindow::set_level_info(int level)
{
  level_label->set_text(String::ucompose("Level: %1", level));
}

void MainWindow::display_message(const Glib::ustring &message, bool centered)
{
  if (!canvas_message.get())
    canvas_message.reset(new CanvasMessage(get_text_layer()));
  
  canvas_message->set_message(message, centered);
  canvas_message->show();
}

void MainWindow::remove_message()
{
  if (canvas_message.get())
    canvas_message->hide();
}

void MainWindow::start_new_game(int start_level, Vector<int> dim)
{
  // make sure no old info is hanging around
  resurrections_label->set_text("");
  player1_status_label->set_text("");
  player2_status_label->set_text("");
  level_label->set_text("");
  
  intro.reset();
  game.reset(); 		// make sure old is gone before creating new
  game.reset(new Game(start_level, dim));

  end_game_item->set_sensitive(true);
  pause_game_item->set_sensitive(true);
  introduction_item->set_sensitive(false);
}

void MainWindow::end_game()
{
  end_game_item->set_sensitive(false);
  pause_game_item->set_sensitive(false);
  introduction_item->set_sensitive(true);
  
  game.reset();

  intro.reset(new Intro);
}

void MainWindow::unpause_game()
{
  MainWindow::instance().remove_message();
  pause_game_item->set_sensitive(true);
}

void MainWindow::generate_background(Vector<int> dim)
{
  Glib::RefPtr<Gdk::Pixbuf> p
    = Gdk::Pixbuf::create(Gdk::COLORSPACE_RGB, false, 8, dim.x, dim.y);

  for (PixelIterator i = begin(p), e = end(p); i != e; ++i) {
#if 1   // FIXME:
    i->red() = 200 + std::rand() % 30 - 15;
    i->blue() = 200 + std::rand() % 30 - 15;
    i->green() = 200 + std::rand() % 30 - 15;
#else
    i->red() = 200;
    i->blue() = 200;
    i->green() = 200;
#endif
  }
  
  background.reset(new Graphic(get_background_layer()));
  background->set_pixbuf(p);

  double x1, y1, x2, y2;
  canvas->get_scroll_region(x1, y1, x2, y2);
  
  background->place_at(make_vector(x1, y1));
}


// call-backs

void MainWindow::on_new_game_activated()
{
  if (new_game_window.get() == 0) {
    new_game_window.reset(new NewGameWindow);
    new_game_window->show();
  }
  else {
    new_game_window->show();
    new_game_window->raise();
  }
}

void MainWindow::on_pause_game_activated()
{
  if (game.get()) {
    if (game->pause()) {
      pause_game_item->set_sensitive(false);
      display_message(_("Game is paused.\n"
			"Press the space bar to continue."));
    }
  }
}

void MainWindow::on_introduction_activated()
{
  assert(intro.get());

  intro->start_over();
}

void MainWindow::on_preferences_activated()
{
  if (preferences_window.get() == 0) {
    preferences_window.reset(new PreferencesWindow);
    preferences_window->show();
  }
  else {
    preferences_window->show();
    preferences_window->raise();
  }
}

void MainWindow::on_quit_activated()
{
  window->hide();
}

void MainWindow::on_about_activated()
{
  if (about.get() == 0) {
    std::vector<Glib::ustring> authors;
    authors.push_back("Ole Laursen <olau@hardworking.dk>");
  
    std::vector<Glib::ustring> documenters;
    // add documenters here

    // note to translators: please fill in your names and email addresses
    Glib::ustring translators = _("translator_credits");

    Glib::ustring description =
      _("Fast-paced monster mashing game for GNOME. Splat the monsters "
	"before they roll over you and take over the world of the gnomes!");
  
    about.reset(new Gnome::UI::About(_("Monster Masher"), VERSION,
				     _("Copyright (c) 2003 Ole Laursen"),
				     authors, documenters, description,
				     translators == "translator_credits"
				     ? "" : translators));
    //, Glib::RefPtr< Gdk::Pixbuf > &logo)
    about->show();
  }
  else {
    about->show();
    about->raise();
  }
}

void MainWindow::on_key_press_event(GdkEventKey* event)
{
  if (intro.get())
    intro->key_pressed(event->keyval);
  else if (game.get())
    game->key_pressed(event->keyval);
}

void MainWindow::on_key_release_event(GdkEventKey* event)
{
  if (game.get())
    game->key_released(event->keyval);
}

bool MainWindow::on_closed(GdkEventAny *)
{
  return false;
}
