/* Main window class.
 *
 * Copyright (c) 2003 Ole Laursen.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

#ifndef MAIN_WINDOW_HPP
#define MAIN_WINDOW_HPP

#include <memory>

#include <libglademm/xml.h>
#include <libgnomeuimm/about.h>
#include <libgnomeuimm/app.h>
#include <libgnomecanvasmm/canvas.h>
#include <libgnomecanvasmm/group.h>

#include <gtkmm/label.h>
#include <gtkmm/menuitem.h>

#include <sigc++/sigc++.h>

#include <gconfmm/client.h>

#include "helpers.hpp"
#include "vector.hpp"


class PreferencesWindow;
class NewGameWindow;
class CanvasMessage;
class Graphic;
class Intro;
class Game;

class MainWindow: public SigC::Object, noncopyable
{
public:
  // MainWindow is a singleton
  static MainWindow &instance();
  ~MainWindow();
  
  void set_canvas_size(Vector<int> dim, Vector<int> offset);
  void set_resurrections(int resurrections);
  void set_player_info(int player_no, int score);
  void set_level_info(int level);

  void display_message(const Glib::ustring &message, bool centered = true);
  void remove_message();
  
  void start_new_game(int start_level, Vector<int> dim);
  void end_game();
  void unpause_game();
  
  Gtk::Window &get_window();

  Gnome::Canvas::Group &get_background_layer();
  Gnome::Canvas::Group &get_block_layer();
  Gnome::Canvas::Group &get_creature_layer();
  Gnome::Canvas::Group &get_text_layer();
  
  Glib::RefPtr<Gnome::Conf::Client> gconf_client;
  const Glib::ustring gconf_dir;

private:
  MainWindow();

  Glib::RefPtr<Gnome::Glade::Xml> ui;

  Gnome::UI::App *window;
  Gnome::Canvas::Canvas *canvas;

  Gtk::MenuItem *new_game_item;
  Gtk::MenuItem *end_game_item;
  Gtk::MenuItem *pause_game_item;
  Gtk::MenuItem *introduction_item;

  Gtk::Label *resurrections_label;
  Gtk::Label *player1_status_label;
  Gtk::Label *player2_status_label;
  Gtk::Label *level_label;

  std::auto_ptr<Graphic> background;
  void generate_background(Vector<int> dim);
  
  // canvas layers
  Gnome::Canvas::Group *background_layer;
  Gnome::Canvas::Group *block_layer;
  Gnome::Canvas::Group *creature_layer;
  Gnome::Canvas::Group *text_layer;
  
  
  // callbacks
  void on_new_game_activated();
  void on_pause_game_activated();
  void on_introduction_activated();
  void on_preferences_activated();
  void on_scores_activated();
  void on_quit_activated();
  void on_about_activated();

  void on_key_press_event(GdkEventKey* event);
  void on_key_release_event(GdkEventKey* event);
  
  bool on_closed(GdkEventAny *);
  
  std::auto_ptr<Gnome::UI::About> about;
  std::auto_ptr<PreferencesWindow> preferences_window;
  std::auto_ptr<NewGameWindow> new_game_window;
  std::auto_ptr<CanvasMessage> canvas_message;
  std::auto_ptr<Intro> intro;
  std::auto_ptr<Game> game;
};

#endif
